with Ada.Text_IO,Messages, Reps.Ops, Reps.IO, Rounding;
use Ada.Text_IO,Messages, Reps.Ops, Reps.IO;

package body Basics is

   Trunc:   constant Boolean := IsNumeric(Scal(Zero));
   ScalOne: constant Scalar  := Scal(One);

   function FirstStep(R: Radius; Gm2: Scalar) return Taylor is
      U: Taylor := CMonom(0,ScalOne);
      T: Taylor;
      K: Natural;
   begin
      for I in 0 .. PDeg/2-1 loop
         K := 2*I;
         T := -Gm2*U-Pow3(R,U);
         SetComponent(K+2,Component(K,T)/Scal((K+4)*(K+2)),U);
      end loop;
      return U;
   end FirstStep;

   function SecondStep(R: Radius; Gm2,D0,D1,S: Scalar) return Taylor is
      U: Taylor := CMonom(0,D0);
      V: Taylor := CMonom(0,ScalOne);
      W: Taylor;
      C: Scalar := Gm2*Exp(Scal(2)*S);
   begin
      SetComponent(1,D1,U);
      V := V-C*Exp(Scal(Two));
      for K in 0 .. PDeg-2 loop
         Prod(R,V,U,W);
         W := W-Pow3(R,U);
         SetComponent(K+2,Component(K,W)/Scal((K+1)*(K+2)),U);
      end loop;
      return U;
   end SecondStep;

   function CFirst(R: Radius; Gm2: Scalar; U: Taylor) return Taylor is
      V: Taylor := CMonom(0,ScalOne);
      T: Taylor := -Gm2*U-Pow3(R,U);
      K: Natural;
   begin
      for I in 0 .. PDeg/2-1 loop
         K := 2*I;
         SetComponent(K+2,Component(K,T)/Scal((K+2)*(K+4)),V);
      end loop;
      SetErrComp(Scal(R*R)*Component(PDeg,T)/Scal((PDeg+2)*(PDeg+4)),V);
      return V;
   end CFirst;

   function CSecond(R: Radius; Gm2,D0,D1,S: Scalar; U: Taylor) return Taylor is
      V: Taylor := CMonom(0,ScalOne);
      W: Taylor;
      C: Scalar := Gm2*Exp(Scal(2)*S);
   begin
      V := V-C*Exp(Scal(Two));
      Prod(R,U,V,W);
      return Dm2(R,D0,D1,W-Pow3(R,U));
   end CSecond;

   function Lip1(R: Radius; Gm2,U: Scalar) return Scalar is
   begin
      return Scal(R*R)*(Gm2+Scal(3)*Sqr(U))/Scal(8);
   end Lip1;

   function Lip2(R: Radius; Gm2,S,U: Scalar) return Scalar is
      SR: Scalar := Sqr(Scal(R));
      ES: Scalar := Scal(One)+Gm2*Exp(Scal(2)*S);
   begin
      return SR*(ES+Scal(3)*Sqr(U))/Scal(Two);
   end Lip2;

   procedure EndPointU(R0: in radius; R,Eps: in Scalar; U: in Taylor;
                      D0,D1: out Scalar) is
      U0,U1: Taylor;
      R1: Radius := Sup(R);     -- radius for the 3rd derivative
   begin
      U0 := Eps*Ball(R0)+U;
      if R1 >= R0 then Message("Basics.EndPointU",Parameter_Error); end if;
      D0 := R*Val(R0,R,U);
      U1 := D(R0,R1,U);         -- derivatives of U
      D1 := D0+R*R*Val(R1,R,U1);
   end EndPointU;

   procedure EndPointW(R0: in radius; R,Eps: in Scalar; U: in Taylor;
                      D0,D1: out Scalar) is
      U0,U1: Taylor;
      R1: Radius := Sup(R);     -- radius for the 3rd derivative
   begin
      U0 := Eps*Ball(R0)+U;
      if R1 >= R0 then Message("Basics.EndPointW",Parameter_Error); end if;
      D0 := Val(R0,R,U);
      U1 := D(R0,R1,U);         -- derivatives of U
      D1 := Val(R1,R,U1);
   end EndPointW;

   function Rad(Gm2,S: Scalar; U: Taylor) return Radius is
      R: Radius := 10.0;
      OneMinus: Radius := 0.9;
   begin
      for I in 1 .. 100 loop
         if Sup(Lip2(R,Gm2,S+Scal(R),Norm(R,U))) < OneMinus then return R; end if;
         R := OneMinus*R;
      end loop;
      Message("Basics.Rad",Giving_Up);
      return Zero;
   end Rad;

   function FindZero(R: Radius; DS: Scalar; U: Taylor) return Scalar is
      A,C: Scalar := Scal(Zero);
      B: Scalar := DS;
      Iter: Positive := 10;
   begin
      for I in 1 .. Iter loop
         C := (A+B)/Scal(2);
         if Sup(Val(R,C,U)) < Zero then
            A := C;
         else
            B := C;
         end if;
      end loop;
      return A;
   end FindZero;

   function Integrate(Gm2: Scalar) return Scalar is
      U,V: Taylor;
      R: Radius := 1.1; -- radius of analyticity
      MaxRepet: constant Natural := 40;
      Eps,D0,D1: Scalar;
      DS,S: Scalar := Scal(Zero);
      NotNeg: Boolean := True;
      MaxL: Radius := 0.95;
   begin
      U := FirstStep(R,Gm2);
      V := U-CFirst(R,Gm2,U);
      SetComponent(0,Scal(Zero),V);  -- resets 1st component, which is 0 by definition
      if Sup(Lip1(R,Gm2,Norm(R,U))) >= MaxL then
         Message("in Basics.Integrate",Domain_Violation);
      end if;
      Eps := Widen(Norm(R,V)/(Scal(One)-Lip1(R,Gm2,Norm(R,U))));
      EndPointU(R,ScalOne,Eps,U,D0,D1);
      while NotNeg or Sup(D0) < Zero loop
         S := S+DS;
         R := Rad(Gm2,S,U);  -- first attempt to compute radius of analyticity
         U := SecondStep(R,Gm2,D0,D1,S);
         R := Rad(Gm2,S,U);        -- reliable radius of analyticity
         DS := Scal(R*0.99);
         V := U-CSecond(R,Gm2,D0,D1,S,U);
         SetComponent(0,Scal(Zero),V);  -- resets components that are 0 by definition
         SetComponent(1,Scal(Zero),V);
         Eps := Norm(R,V);
         Eps := Widen(Eps/(Scal(One)-Lip2(R,Gm2,S+DS,Eps+Norm(R,U))));

         EndPointW(R,DS,Eps,U,D0,D1);
         NotNeg := NotNeg and Sup(D0) >= Zero;  -- false as soon as D0<0
      end loop;
      DS := FindZero(R,DS,U);
      if Sup(Val(R,DS,U)) > Zero then
         Show("Something is wrong");
         return Scal(Zero);
      end if;
      return Exp(S+DS);
   end Integrate;

   function Check(MinP,MaxP: Rep; File: String) return Boolean is
      Gamma,Z: Scalar;
      SqrtMu1Lower: Rep := 3.831705970;
      SqrtMu1Upper: Rep := 3.831705971;
      MinGamma: Rep := MinP/SqrtMu1Upper;
      MaxGamma: Rep := MaxP/SqrtMu1Lower;
      G1,G2: Rep := MinGamma;
      RelWidth: Rep := 1.00005;
      SqrtMu1: Scalar := Scal(SqrtMu1Lower,SqrtMu1Upper);
      Gm2: Scalar;
      F: File_Type;
      OK: Boolean := True;
   begin
      Create(F,Out_File,File);
      while G2 < MaxGamma loop
         G2 := G1*RelWidth;
         Gamma := Scal(G1,G2);
         Gm2 := Scal(One)/(Gamma*Gamma);
         Txt_Put(F,G1);
         Z := Integrate(Gm2);
         Txt_Put(F,Inf(Z));
         Txt_Put(F,Inf(Z-Gamma*SqrtMu1));
         OK := OK and Inf(Z-Gamma*SqrtMu1) > Zero;
         New_Line(F);
         G1 := G2;
      end loop;
      Close(F);
      return OK;
   end Check;

end Basics;
